/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.controller;

import com.funambol.blackberry.BBVersion;
import com.funambol.blackberry.Configuration;
import com.funambol.blackberry.Customization;
import com.funambol.blackberry.SyncClient;
import com.funambol.blackberry.SyncEngine;
import com.funambol.blackberry.ui.DisplayManager;
import com.funambol.util.ThreadPool;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import com.funambol.util.AppProperties;
import com.funambol.util.StaticDataHelper;
import com.funambol.util.TransportAgent;
import com.funambol.updater.Updater;
import com.funambol.updater.BasicUpdaterConfig;
import com.funambol.updater.UpdaterListener;

import net.rim.blackberry.api.browser.Browser;

import java.util.Date;

/**
 * This class is the client interface to the automatic update library.
 */
public class UpdaterManager implements UpdaterListener {

    /** Default timing definition */
    public static final long DEFAULT_CHECK_INTERVAL = Customization.getInstance().getCheckUpdtIntervalDefault();
    public static final long DEFAULT_REMINDER_INTERVAL = Customization.getInstance().getReminderUpdtIntervalDefault();

    /**
     * The BlackBerry Sync Client component name:
     *  - bbpim for BlackBerry OS from 4.2.1 to 4.6
     *  - bbpim_os47 for BlackBerry OS 4.7
     */
    private String componentName           = "bbpim";

    private Updater updater                = null;

    /** The update configuration */
    private UpdaterManagerConfig config    = null;

    /** Update procedure timing definition */
    private long checkInterval = DEFAULT_CHECK_INTERVAL;
    private long reminderInterval = DEFAULT_REMINDER_INTERVAL;

    /** The UpdaterManager singleton */
    private static UpdaterManager instance = null;

    private Controller controller;

    private final int UPDATE_NOW_VALUE = 0;
    private final int UPDATE_LATER_VALUE = 1;
    private final int SKIP_UPDATE_VALUE = 2;

    private final String UPDATE_NOW_LABEL = StaticDataHelper.getLanguage("update_command-now");
    private final String UPDATE_LATER_LABEL = StaticDataHelper.getLanguage("update_command-later");
    private final String SKIP_UPDATE_LABEL = StaticDataHelper.getLanguage("update_command-skip");

    private final String UPDATE_POPUP_MESSAGE = StaticDataHelper.getLanguage("message_update-available");
    
    /**
     * Get the singleton of the <code>UpdaterManager</code>
     * @return The <code>UpdaterManager</code> singleton
     */
    public static UpdaterManager getInstance() {
        if (instance == null) {
            instance = new UpdaterManager();
        }
        return instance;
    }

    /**
     * Set the client <code>Controller</code> reference.
     * @param controller The <code>Controller</code> reference
     */
    public void setController(Controller controller){
        this.controller = controller;
    }

    /**
     * The <code>Thread</code> used to run the Update procedure
     */
    private class UpdaterThread extends Thread {
        
        public UpdaterThread() { }

        public void run() {
            updater.check();
        }
    }

    /**
     * Manages the Updater configuration parameters.
     */
    private class UpdaterManagerConfig extends BasicUpdaterConfig {

        /**
         * Load the Update manager configuration
         */
        public void load() {
            try {
                Configuration.load();
            } catch (Exception ce) {
                Log.error("[UpdaterManagerConfig.load]Loading updater config " +
                        "failed: " + ce);

                // We shall create a new config
                String server = Customization.getInstance().getServerUriDefault();
                server = StringUtil.extractAddressFromUrl(server);

                Log.debug("[UpdaterManagerConfig.load]Server url: " + server);

                super.setUrl(server);
                checkInterval = Customization.getInstance().getCheckUpdtIntervalDefault();
                setCheckInterval(checkInterval);
                reminderInterval = Customization.getInstance().getReminderUpdtIntervalDefault();
                setReminderInterval(reminderInterval);
            }
            String server = Configuration.getSyncUrl();
            server = StringUtil.extractAddressFromUrl(server);
            Log.debug("[UpdaterManagerConfig.load]Server url: " + server);
            setUrl(server);
            setDownloadUrl(Configuration.getDownloadUrl());
            setType(Configuration.getUpdateType());
            setAvailableVersion(Configuration.getAvailableVersion());
            setLastCheck(Configuration.getLastUpdateCheck());
            setCheckInterval(Configuration.getCheckInterval());
            setReminderInterval(Configuration.getReminderInterval());
            setLastReminder(Configuration.getLastReminder());
            setSkip(Configuration.getSkip());
        }

        /**
         * Save the Update manager configuration
         */
        public void save() {
            try {
                Configuration.setDownloadUrl(getDownloadUrl());
                Configuration.setUpdateType(getType());
                Configuration.setAvailableVersion(getAvailableVersion());
                Configuration.setLastUpdateCheck(getLastCheck());
                Configuration.setCheckInterval(getCheckInterval());
                Configuration.setReminderInterval(getReminderInterval());
                Configuration.setLastReminder(getLastReminder());
                Configuration.setSkip(getSkip());
                Configuration.saveConfiguration();
            } catch (Exception ce) {
                Log.error("[UpdaterManagerConfig.save]Saving updater config failed " + ce);
            }
        }
    }

    /**
     * Default constructor
     */
    private UpdaterManager() {
        String version = BBVersion.VERSION;
        config = new UpdaterManagerConfig();
        config.load();
        String [] osVersion = StaticDataHelper.getOSVersion();
        int os = Integer.parseInt(osVersion[0] + osVersion[1]);
        if(os > 46){
            componentName = componentName + "_os47";
        }
        updater = new Updater(config, version, componentName);
        updater.setListener(this);
    }

    /**
     * Check for updates
     */
    public void check() {
        // The update check is executed in a separate thread
        UpdaterThread updThread = new UpdaterThread();
        updThread.setPriority(Thread.MIN_PRIORITY);
        updThread.start();
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void optionalUpdateAvailable(String version) {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            DialogController dialControll = controller.getDialogController();
            int selection = dialControll.promptSelection(UPDATE_POPUP_MESSAGE,
                                                    new String[]{UPDATE_NOW_LABEL,UPDATE_LATER_LABEL,SKIP_UPDATE_LABEL},
                                                    new int[]{UPDATE_NOW_VALUE,UPDATE_LATER_VALUE,SKIP_UPDATE_VALUE},
                                                    UPDATE_NOW_VALUE,
                                                    true,
                                                    DisplayManager.NO_LIMIT);
            if (selection == UPDATE_NOW_VALUE){
                this.userSelectedNow();
            }else if (selection == UPDATE_LATER_VALUE){
                this.userSelectedLater();
            }else{
                //selected SKIP_UPDATE_VALUE
                this.userSelectedSkip();
            }
        }
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void recommendedUpdateAvailable(String version) {
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            DialogController dialControll = controller.getDialogController();
            int selection = dialControll.promptSelection(UPDATE_POPUP_MESSAGE,
                                                    new String[]{UPDATE_NOW_LABEL,UPDATE_LATER_LABEL},
                                                    new int[]{UPDATE_NOW_VALUE,UPDATE_LATER_VALUE},
                                                    UPDATE_NOW_VALUE,
                                                    true,
                                                    DisplayManager.NO_LIMIT);
            if (selection == UPDATE_NOW_VALUE){
                this.userSelectedNow();
            }else{
                //selected UPDATE_LATER_VALUE
                this.userSelectedLater();
            }
        }
    }

    /**
     * @see com.funambol.updater.UpdaterListener
     */
    public void mandatoryUpdateAvailable(String version) {
        // TODO: Actually we don't support mandatory updates treat it as
        //       recommended update.
        Log.debug("[UpdaterManager.mandatoryUpdateAvailable] Mandatory update" +
                " not supported, treat it as recommended.");
        recommendedUpdateAvailable(version);
    }

    /**
     * Update the last update reminder timestamp.
     */
    private void updateLastReminder() {
        long now = System.currentTimeMillis();
        Log.info("UpdaterManager - a reminder will be showed after: " +
                 new Date(now+ config.getReminderInterval()));
        updater.setLastReminder(now);
        config.save();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Now" to
     * the update message.
     */
    public void userSelectedNow() {
        Log.info("UpdaterManager - update now");
        this.update();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Later" to
     * the update message.
     */
    public void userSelectedLater() {
        Log.info("UpdaterManager - update later");
        this.updateLastReminder();
    }

    /**
     * Tells the <code>UpdaterManager</code> that the user choosed "Skip" to
     * the update message.
     */
    public void userSelectedSkip() {
        Log.info("UpdaterManager - skip update");
        updater.setSkip();
    }

    /**
     * Check wheter an update is currently available.
     * @return
     */
    public boolean isUpdateAvailable() {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            return updater.isUpdateAvailable();
        } else {
            return false;
        }
    }

    /**
     * Get the url where the new client version is located.
     * @return The update download url.
     */
    public String getDownloadUrl() {
        return config.getDownloadUrl();
    }

    /**
     * Set the url where the new client version is located.
     * @param url The update download url.
     */
    public void setUrl(String url) {
        if(!url.equals(config.getUrl())) {
            config.setUrl(url);
            // Since a new url has been set, we also reset our timers to force a new
            // check and possibly a new reminder
            config.setLastReminder(0);
            config.setLastCheck(0);
            config.save();
        }
    }

    /**
     * Update the client. It Opens the default Browser to the update download
     * location.
     */
    public void update(){
        Log.info("UpdaterManager - updating client from: " + config.getDownloadUrl());
        Browser.getDefaultSession().displayPage(config.getDownloadUrl());
    }
}

